from __future__ import unicode_literals

# Basic extensions to the python language or standard library.
# As a rule of thumb, objects and functions here shouldn't depend
# on any other modules from package.
